from datetime import datetime, timedelta


class EventObserver:
    _last_event: datetime
    _now: callable

    def __init__(self, min_delay_millis, time_provider=None):
        self._min_delay_millis = min_delay_millis
        if time_provider is None:
            time_provider = datetime.utcnow
        self._now = time_provider
        self._update()

    def _update(self):
        self._last_event = self._now()

    def event_happened(self):
        self._update()

    def is_stable(self):
        if self._min_delay_millis == 0:
            return True
        now = self._now()
        diff: timedelta = now - self._last_event
        return diff.total_seconds() * 1000 > self._min_delay_millis



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from functools import partial
from typing import Optional, Tuple, List

import js
import wwwpy.remote.component as wpc
from js import document, console, Event, HTMLElement, window
from pyodide.ffi import create_proxy
from wwwpy.common import state, modlib
from wwwpy.common.designer import element_library, new_component
from wwwpy.common.designer.code_edit import add_component, ElementDef
from wwwpy.common.designer.element_library import Help
from wwwpy.common.designer.element_path import ElementPath
from wwwpy.common.designer.html_edit import Position
from wwwpy.common.designer.html_locator import path_to_index
import asyncio
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer import element_path
from wwwpy.remote.designer.drop_zone import DropZone
from wwwpy.remote.designer.global_interceptor import GlobalInterceptor, InterceptorEvent
from wwwpy.remote.designer.ui.draggable_component import DraggableComponent
from wwwpy.server.designer import rpc

from wwwpy.remote.designer.helpers import _element_lbl, _help_button, info_link, _help_url
from wwwpy.remote.designer.ui.property_editor import PropertyEditor
from . import filesystem_tree
from .help_icon import HelpIcon  # noqa
import logging

logger = logging.getLogger(__name__)


@dataclass
class ToolboxState:
    geometry: Tuple[int, int, int, int] = field(default=(200, 30, 400, 330))
    toolbox_search: str = ''
    selected_element_path: Optional[ElementPath] = None


@dataclass
class MenuMeta:
    label: str
    html: str
    always_visible: bool = False
    p_element: js.HTMLElement = None
    help: Help = None


def menu(label, always_visible=False):
    def wrapped(fn, label=label):
        help = None
        if isinstance(label, Help):
            help = label
            label = label.description
        fn.label = label
        fn.meta = MenuMeta(label, label, always_visible, help=help)
        return fn

    return wrapped


class ToolboxComponent(wpc.Component, tag_name='wwwpy-toolbox'):
    body: HTMLElement = wpc.element()
    inputSearch: js.HTMLInputElement = wpc.element()
    dragComp1: DraggableComponent = wpc.element()
    property_editor: PropertyEditor = wpc.element()
    _select_element_btn: js.HTMLElement = wpc.element()
    _select_clear_btn: js.HTMLElement = wpc.element()
    components_marker = '-components-'

    @property
    def visible(self) -> bool:
        return self.dragComp1.element.style.display != 'none'

    @visible.setter
    def visible(self, value: bool):
        self.dragComp1.element.style.display = 'block' if value else 'none'

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<style>
.two-column-layout {
  display: flex;
  flex-wrap: wrap;
}

.two-column-layout p {
  width: calc(50% - 10px); /* 50% width minus half of the gap */
  margin: 0 20px 10px 0; /* Right margin creates gap between columns */
}

.two-column-layout p:nth-child(even) {
  margin-right: 0; /* Removes right margin for every even child */
}
</style>     
<wwwpy-draggable-component data-name='dragComp1'>
    <div slot='title' style='text-align: center'>wwwpy</div>
    <div  style="text-align: center; padding: 8px">     
        <button data-name="_select_element_btn">Select element...</button>   
        <button data-name="_select_clear_btn">Clear selection</button>   
        <wwwpy-property-editor data-name="property_editor"></wwwpy-property-editor>        
        <p><input data-name='inputSearch' type='search' placeholder='type to filter...'></p>
        <div data-name='body' class='two-column-layout'></div>
    </div>   
</wwwpy-draggable-component>         
"""
        self._manage_toolbox_state()

        attrs = [v for k, v in vars(self.__class__).items() if hasattr(v, 'label')]
        self._all_items: List[MenuMeta] = []

        def add_p(menu_meta: MenuMeta, callback):  # callback can be async
            self._all_items.append(menu_meta)
            p: js.HTMLElement = document.createElement('p')
            p.style.color = 'white'
            menu_meta.p_element = p
            p.innerHTML = menu_meta.html
            p.addEventListener('click', create_proxy(callback))

        def add_comp(element_def: ElementDef):

            def _on_pointed(drop_zone: DropZone | None):
                console.log(f'pointed dropzone: {drop_zone}')
                msg = (f'Insert new {element_def.tag_name}')
                if drop_zone:
                    pos = 'before' if drop_zone.position == Position.beforebegin else 'after'
                    # msg += f' at {drop_zone.position.name} of {drop_zone.element.tagName}'
                    msg += f' {pos} {drop_zone.element.tagName}'
                else:
                    msg += ' ... select a dropzone.'
                self.property_editor.message1div.innerHTML = msg

            async def _start_drop_for_comp(event):
                _on_pointed(None)
                res = await _drop_zone_start_selection_async(_on_pointed)
                logger.debug(f'_start_drop_for_comp res={res}')
                if res:
                    await self._process_dropzone(res, element_def)
                else:
                    self.property_editor.message1div.innerHTML = 'canceled'

            element_html = f'<span style="cursor: pointer">{element_def.tag_name}</span> {_help_button(element_def)}'
            add_p(MenuMeta(element_def.tag_name, element_html), _start_drop_for_comp)

        for member in attrs:
            if member.meta.label == self.components_marker:
                [add_comp(ele_def) for ele_def in element_library.element_library().elements]
            else:
                help = member.meta.help
                if help:
                    help_html = '' if not help.url else info_link(help.url)
                    element_html = f'{help.description} {help_html}'
                    add_p(MenuMeta(help.description, element_html), partial(member, self))
                else:
                    add_p(member.meta, partial(member, self))
        self._update_toolbox_elements()

    async def _process_dropzone(self, drop_zone: DropZone, element_def: ElementDef):
        el_path = element_path.element_path(drop_zone.element)

        if not el_path:
            window.alert(f'No component found for dropzone!')
            return

        logger.debug(f'element_path={el_path}')
        file = modlib._find_module_path(el_path.class_module)
        old_source = file.read_text()

        path_index = path_to_index(el_path.path)
        add_result = add_component(old_source, el_path.class_name, element_def, path_index, drop_zone.position)

        if add_result:
            logger.debug(f'write_module_file len={len(add_result.html)} el_path={el_path}')
            new_element_path = ElementPath(el_path.class_module, el_path.class_name, add_result.node_path, el_path.origin)
            self._toolbox_state.selected_element_path = new_element_path
            write_res = await rpc.write_module_file(el_path.class_module, add_result.html)
            logger.debug(f'write_module_file res={write_res}')

    def _manage_toolbox_state(self):
        self._toolbox_state = state._restore(ToolboxState)
        self.inputSearch.value = self._toolbox_state.toolbox_search
        self.dragComp1.set_geometry(self._toolbox_state.geometry)

        def on_toolbar_geometry_change():
            if self.dragComp1.acceptable_geometry():
                self._toolbox_state.geometry = self.dragComp1.geometry()

        self.dragComp1.geometry_change_listeners.append(on_toolbar_geometry_change)
        self._restore_selected_element_path()

    def inputSearch__input(self, e: Event):
        self._toolbox_state.toolbox_search = self.inputSearch.value
        self._update_toolbox_elements()

    def _update_toolbox_elements(self):
        self.body.innerHTML = ''
        for meta in self._all_items:
            p = meta.p_element
            if meta.always_visible or self.inputSearch.value.lower() in meta.label.lower():
                self.body.appendChild(p)

    async def _select_clear_btn__click(self, e: Event):
        self._toolbox_state.selected_element_path = None
        self._restore_selected_element_path()

    async def _select_element_btn__click(self, e: Event):
        no_comp = 'Click an element...'
        self.property_editor.message1div.innerHTML = no_comp

        def _on_pointed(drop_zone: DropZone | None):
            msg = no_comp
            if drop_zone:
                msg = 'Click to select ' + _element_lbl(drop_zone.element)
            self.property_editor.message1div.innerHTML = msg
            console.log(f'pointed dropzone: {drop_zone}')

        res = await _drop_zone_start_selection_async(_on_pointed, whole=True)
        self._toolbox_state.selected_element_path = element_path.element_path(res.element) if res else None
        self._restore_selected_element_path()

    @menu(components_marker)
    def _drop_zone_start(self, e: Event):
        assert False, 'Just a placeholder'

    @menu(Help('Create new component', _help_url('add_component')))
    async def _add_new_component(self, e: Event):
        if js.window.confirm('Add new component file?\nIt will be added to your "remote" folder.'):
            res = await rpc.add_new_component()
            js.window.alert(res)

    @menu(Help("Explore local filesystem", _help_url('remote_filesystem')))
    async def _browse_local_filesystem(self, e: Event):
        filesystem_tree.show_explorer()

    # @menu('handle global click')
    def _handle_global_click(self, e: Event):
        # add input
        def global_click(event: Event):
            event.preventDefault()
            event.stopImmediatePropagation()
            event.stopPropagation()
            console.log('global click', event.element)
            if self._global_click:
                document.removeEventListener('click', self._global_click, True)
                self._global_click = None

        self._global_click = create_proxy(global_click)
        document.addEventListener('click', self._global_click, True)

    def _restore_selected_element_path(self):
        element_path = self._toolbox_state.selected_element_path
        if element_path:
            console.log(f'restoring selected element path: {element_path}')
            if not element_path.valid():
                element_path = None

        self.property_editor.selected_element_path = element_path
        if element_path:
            self._select_clear_btn.removeAttribute('disabled')
        else:
            self._select_clear_btn.setAttribute('disabled', '')


def is_inside_toolbar(element: HTMLElement | None):
    if not element:
        return False

    orig = element
    # loop until the root element and see if it is the toolbar
    res = False
    while element:
        if element.tagName.lower() == ToolboxComponent.component_metadata.tag_name:
            res = True
            break
        element = element.parentElement

    return res


def _default_drop_zone_accept(drop_zone: DropZone):
    name = drop_zone.element.tagName.lower()
    from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
    accept = not (name == 'body' or name == 'html' or name == DevModeComponent.component_metadata.tag_name)
    return accept


async def _drop_zone_start_selection_async(on_pointed, whole=False) -> Optional[DropZone]:
    from wwwpy.remote.designer.drop_zone import drop_zone_selector

    event = asyncio.Event()
    result = []

    def intercept_ended(ev: InterceptorEvent):
        selected = drop_zone_selector.stop()
        if is_inside_toolbar(ev.target):
            ev.uninstall()
            event.set()
            return
        ev.preventAndStop()
        if selected:
            ev.uninstall()
            console.log(
                f'selection accepted position {selected.position.name} target: ', selected.element,
                'parent: ', selected.element.parentElement,
                'event: ', ev.event,
                'composedPath: ', ev.event.composedPath(),
            )
            result.append(selected)
            event.set()

    GlobalInterceptor(intercept_ended, 'pointerdown').install()

    click_inter = GlobalInterceptor(lambda ev: ev.preventAndStop(), 'click')
    click_inter.install()
    drop_zone_selector.start_selector(on_pointed, _default_drop_zone_accept, whole=whole)
    await event.wait()
    await asyncio.sleep(0.5)
    click_inter.uninstall()

    if len(result) == 0:
        return None

    return result[0]



import base64
import mimetypes

from js import console, document
from pathlib import Path


def download_path(filename: str, path: Path, mime_type: str = None):
    download_bytes(filename, path.read_bytes(), mime_type)


def download_bytes(filename: str, content: bytes, mime_type: str = None):
    if mime_type is None:
        gt = mimetypes.guess_type(filename)
        mime_type = gt[0]
        if mime_type is None:
            mime_type = 'application/octet-stream'
        console.log(f'guess mime type for `{filename}` is `{mime_type}`')
    a = document.createElement('a')
    a.download = filename
    a.href = f'data:{mime_type};base64,{base64.b64encode(content).decode("ascii")}'
    document.body.append(a)
    a.click()




from __future__ import annotations

import inspect
import sys


def reload(module):
    import importlib
    # importlib.invalidate_caches()
    return importlib.reload(module)


def unload_path(path: str):
    def accept(module):
        try:
            module_path = inspect.getfile(module)
            return module_path.startswith(path) and module_path != __file__
        except:
            return False

    names = [name for name, module in sys.modules.items() if accept(module)]

    for name in names:
        print(f'reloader - Deleting module {name}...')
        del (sys.modules[name])



import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List, Any, Optional

from wwwpy.common.rpc import serialization
from wwwpy.common.filesystem.sync import Event


@dataclass
class Operation:
    op_type: str
    is_dir: bool
    path: str
    was: str = ""
    content_bytes: Optional[bytes] = None
    content_str: Optional[str] = None


def sync_source(source: Path, events: List[Event]) -> List[Any]:
    res = _make_summary(source, events)
    ser = serialization.serialize(res, List[Operation])
    return ser


def _make_summary(source: Path, events: List[Event]) -> List[Operation]:
    state = {}
    moved = {}
    for e in events:
        src = Path(e.src_path)
        name = str(src.relative_to(source))
        prev = state.get(name, None)
        if e.is_directory:
            if e.event_type == 'deleted':
                for name_p in list(state.keys()):
                    if name_p.startswith(name):
                        del state[name_p]
                state[name] = 'deleted'

        else:  # is file
            if e.event_type == 'moved':
                dst = Path(e.dest_path)
                dst_name = str(dst.relative_to(source))
                moved[dst_name] = name
            if e.event_type == 'created':
                state[name] = 'modified'
            if e.event_type == 'modified':
                state[name] = state.get(name, 'modified')  # if it was created, it stays created
            elif e.event_type == 'deleted':
                if prev == 'modified':
                    del state[name]
                else:
                    state[name] = 'deleted'

    result = []
    for name, status in state.items():
        path = source / name
        if status == 'deleted':
            result.append(Operation(op_type='delete', is_dir=False, path=str(name)))
        elif status == 'modified':
            _append_file(result, name, path)
    return result


def _append_file(result, name, path):
    if path.is_file() and path.exists():
        try:
            result.append(Operation(op_type='write', is_dir=False, path=str(name), content_str=path.read_text()))
            # result.append({'name': str(name), 'content': path.read_text()})
        except UnicodeDecodeError:
            result.append(Operation(op_type='write', is_dir=False, path=str(name), content_bytes=path.read_bytes()))
        except Exception as e:
            pass


def sync_target(target_root: Path, changes: List[Any]) -> None:
    res = serialization.deserialize(changes, List[Operation])
    _apply_summary(target_root, res)


def _apply_summary(target_root: Path, changes: List[Operation]) -> None:
    for change in changes:
        target = target_root / change.path
        if change.op_type == 'delete':
            if target.exists():
                if target.is_dir():
                    shutil.rmtree(target)
                else:
                    target.unlink(missing_ok=True)
        elif change.op_type == 'write':

            content = change.content_str
            if content is None:
                content = change.content_bytes

            if content is not None:
                target.parent.mkdir(parents=True, exist_ok=True)
                if isinstance(content, bytes):
                    target.write_bytes(content)
                else:
                    target.write_text(content)
            else:
                if target.exists():
                    if target.is_dir():
                        shutil.rmtree(target)
                    else:
                        target.unlink(missing_ok=True)


def sync_init(source: Path) -> List[Any]:
    summary = _make_initial_summary(source)
    serialize = serialization.serialize(summary, List[Operation])
    return serialize


def _make_initial_summary(source: Path) -> List[Operation]:
    result = []
    for path in source.rglob('*'):
        _append_file(result, path.relative_to(source), path)
    return result



from pathlib import Path
from threading import Thread
from time import sleep
from typing import Callable

from watchdog.events import FileSystemEventHandler, FileSystemEvent
from watchdog.observers import Observer

from wwwpy.common.filesystem.sync import new_tmp_path

import logging

logger = logging.getLogger(__name__)


class AnyObserver(FileSystemEventHandler):

    def __init__(self, path: Path, callback: Callable[[FileSystemEvent], None]):
        """path need to exist, otherwise the observer will throw an exception."""
        self._path = path
        self._callback = callback
        self._observer = None
        super().__init__()

    def watch_directory(self):
        if self._path.exists():
            self._new_observer()
        else:
            logger.debug(f'path does not exist, waiting for it to be created: {self._path}')
            Thread(target=self._wait_for_path, daemon=True).start()

    @property
    def active(self) -> bool:
        return self._observer is not None and self._observer.is_alive()

    def _wait_for_path(self):
        while not self._path.exists():
            sleep(0.1)
        self._new_observer()

    def _new_observer(self):
        logger.debug(f'Creating observer for {self._path}')
        self._observer = Observer()
        self._observer.schedule(self, str(self._path), recursive=True)
        self._observer.start()

    def stop(self):
        self._observer.stop()

    def join(self):
        try:
            self._observer.join()
        except RuntimeError:
            pass  # catch if it was not started

    def on_any_event(self, event: FileSystemEvent) -> None:
        if not Path(event.src_path).is_relative_to(self._path):
            return
        if self._is_self_delete(event):
            self._observer.stop()
            self._observer = None
            self._wait_for_path()
            return

        self._callback(event)

    def _is_self_delete(self, event: FileSystemEvent) -> bool:
        return event.event_type == 'deleted' and Path(event.src_path) == self._path


def main():
    tmp_path = new_tmp_path()
    print(f'watching {tmp_path}')

    def print_event(event: FileSystemEvent):
        print(f'{event}')

    AnyObserver(tmp_path, print_event).watch_directory()

    while True:
        sleep(1)


if __name__ == '__main__':
    main()



import base64
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


async def upload_init(name: str, size: int):
    file = _resolve_file(name)
    file.unlink(missing_ok=True)
    file.touch()
    logger.info(f'upload_init name={name} size={size}')
    return None


async def upload_append(name: str, b64str: str):
    logger.info(f'upload_append name={name} len(b64str)={len(b64str)}')
    file = _resolve_file(name)
    bytes_append = base64.b64decode(b64str)
    with file.open('ab') as f:
        f.write(bytes_append)
    return None


def _resolve_file(name: str) -> Path:
    folder = Path(__file__).parent.parent / 'uploads'
    candidate = folder / name
    # security check: candidate is inside the folder?
    if not candidate.resolve().is_relative_to(folder.resolve()):
        raise ValueError(f'Invalid path: {candidate}')
    candidate.parent.mkdir(parents=True, exist_ok=True)
    return candidate



from __future__ import annotations

from typing import Callable, TypeVar, Any, Generic, Collection

T = TypeVar('T')
K = TypeVar('K')


class ListMap(list[T]):
    def __init__(self, args: Collection[T] = (), key_func: Callable[[T], K] = None):
        super().__init__(args)
        if key_func is not None:
            self._key = key_func
        self._map = {self._key(item): item for item in self}

    # append = _modify_method(list.append)
    # extend = _modify_method(list.extend, takes_list=True)
    # insert = _modify_method(list.insert, 1)
    # __add__ = _modify_method(list.__add__, takes_list=True)
    # __iadd__ = _modify_method(list.__iadd__, takes_list=True)
    # __setitem__ = _modify_method(list.__setitem__, 1)
    def _key(self, item: T) -> K:
        return item

    def append(self, value: T):
        self._map[self._key(value)] = value
        super().append(value)

    def get(self, key: K) -> T | None:
        """Return the item with the given key or None if it does not exist."""
        return self._map.get(key)



from __future__ import annotations

import logging
import shutil
from dataclasses import dataclass
from pathlib import Path

from wwwpy.common.designer.element_library import NamedListMap
from wwwpy.common.tree import print_tree

_known = {'common', 'remote', 'server'}


logger = logging.getLogger(__name__)


def setup_quickstart(directory: Path, quickstart_name: str):
    assert is_empty_project(directory), f'project is not empty: {directory}'
    source = Path(__file__).parent / quickstart_name
    assert source.exists(), f'quickstart not found: {source}'
    # _delete_empty_project(directory)
    shutil.copytree(source, directory, dirs_exist_ok=True)
    logger.warning(f'Quickstart applied {quickstart_name} to {directory}')
    # print_tree(directory, logger.warning)

def _delete_empty_project(directory: Path):
    assert is_empty_project(directory), f'project is not empty: {directory}'
    for p in _known:
        sub = directory / p
        if sub.exists():
            try:
                shutil.rmtree(sub)
            except Exception as e:
                pass


# def _check_quickstart(directory: Path):
#     if _do_quickstart(directory):
#         print(f'empty directory, creating quickstart in {directory.absolute()}')
#         _setup_quickstart(directory)
#     else:
#         (directory / 'remote').mkdir(exist_ok=True)


# def _do_quickstart(directory):
#     items = list(directory.iterdir())
#     return all(item.name.startswith('.') for item in items)

def _make_hotreload_work(root_path: Path):
    if not is_empty_project(root_path):
        return

    def _write_empty_package(name: str):
        package = root_path / name
        package.mkdir(exist_ok=True)
        (package / '__init__.py').write_text('')

    for p in _known:
        _write_empty_package(p)


@dataclass
class Quickstart:
    name: str
    title: str
    description: str
    path: Path


def quickstart_list() -> NamedListMap[Quickstart]:
    source = Path(__file__).parent
    # keep only files that contains a file called 'readme.txt'
    quickstarts = []

    def add(name: str, readme: Path):
        if not readme.exists():
            return
        lines = readme.read_text().splitlines()
        title = lines[0]
        description = '\n'.join(lines[2:])
        quickstart = Quickstart(name, title, description, source / name)
        quickstarts.append(quickstart)

    for d in source.iterdir():
        if d.is_dir() and not d.name.startswith('.'):
            add(d.name, d / 'readme.txt')
    quickstarts = sorted(quickstarts, key=lambda x: x.name)
    return NamedListMap(quickstarts)


def is_empty_project(root_path: Path | str):
    root_path = Path(root_path)
    from wwwpy.common import files

    def blacklisted(item: Path) -> bool:
        return item.name.startswith('.') or item.name in files.directory_blacklist

    def empty(package: str) -> bool:
        directory = root_path / package
        if not directory.exists():
            return True
        init = directory / '__init__.py'

        if init.exists() and init.read_text().strip() != '':
            return False

        def is_init(item: Path) -> bool:
            return item.name == '__init__.py'

        return all(blacklisted(item) or is_init(item) for item in directory.iterdir())

    known_empty = all(empty(folder) for folder in _known)
    rest_ok = all(i.name in _known or blacklisted(i) for i in root_path.iterdir())
    result = known_empty and rest_ok
    return result



from threading import Thread
from typing import Optional

import wwwpy
from flask import Flask, Response
from wwwpy import Webserver
from wwwpy.routes import Route


class WsFlask(Webserver):
    def __init__(self):
        super().__init__()
        self.app = Flask(__name__)
        self.thread: Optional[Thread] = None

    def _setup_route(self, route: Route):
        def func() -> Response:
            return self.to_native_response(route.callback())

        self.app.add_url_rule(route.path, route.path, func)

    def to_native_response(self, pn_response: wwwpy.HttpResponse) -> Response:
        return Response(pn_response.content, status=200, content_type=pn_response.content_type)

    def _start_listen(self):
        def flask_run():
            self.app.run(host=self.host, port=self.port)

        self.thread = Thread(target=flask_run, daemon=True)
        self.thread.start()



from typing import List, Any

from wwwpy.common import modlib


class DesignerRpc:

    def package_file_changed_sync(self, package_name: str, events: List[Any]):
        from wwwpy.common.filesystem.sync import sync_delta2
        from wwwpy.common.filesystem.sync import Sync
        sync_impl: Sync = sync_delta2
        directory = modlib._find_package_directory(package_name)
        sync_impl.sync_target(directory, events)

        from wwwpy.remote.browser_main import _reload
        _reload()



from pathlib import Path

from wwwpy.common.filesystem.sync import Event


def event_apply(path: Path, event: Event) -> Path:
    src_full_path = path / event.src_path
    dest_full_path = path / event.dest_path if event.dest_path else None

    if event.event_type == 'created':
        if event.is_directory:
            src_full_path.mkdir(exist_ok=True)
        else:
            src_full_path.touch()

    elif event.event_type == 'deleted':
        if event.is_directory:
            src_full_path.rmdir()
        else:
            src_full_path.unlink()

    elif event.event_type == 'modified':
        if not event.is_directory:
            # For simplicity, we'll just update the file's content
            # In a real scenario, you might want to store the old content somewhere
            src_full_path.write_text(f"Modified content at {src_full_path}")

    elif event.event_type == 'moved':
        src_full_path.rename(dest_full_path)

    # For 'closed' event, we don't need to do anything

    return path


from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

import js
from js import Event, document, HTMLElement
from pyodide.ffi import create_proxy


@dataclass
class InterceptorEvent:
    target: HTMLElement | None
    event: Event
    interceptor: GlobalInterceptor

    def uninstall(self):
        self.interceptor.uninstall()

    def preventAndStop(self):
        if self.event:
            self.event.preventDefault()
            self.event.stopImmediatePropagation()
            self.event.stopPropagation()


class GlobalInterceptor:

    def __init__(self, callback: Callable[[InterceptorEvent], None], event_name: str = 'click'):
        self._callback = callback
        self._event_name = event_name
        self._installed = False
        self._proxy = create_proxy(self._handler)

    def install(self):
        if self._installed:
            return
        self._installed = True
        js.document.addEventListener(self._event_name, self._proxy, True)

    def uninstall(self):
        if not self._installed:
            return
        self._installed = False
        js.document.removeEventListener(self._event_name, self._proxy, True)

    def _handler(self, event: Event):
        ev = InterceptorEvent(event.target, event, self)
        self._callback(ev)


def global_interceptor_start(callback: Callable[[InterceptorEvent], None]):
    proxy = []

    def _uninstall():
        document.removeEventListener('click', proxy[0], True)
        proxy.clear()

    def global_click(event: Event):
        ev = InterceptorEvent(event.target, event, _uninstall)
        callback(ev)

    proxy.append(create_proxy(global_click))
    document.addEventListener('click', proxy[0], True)



from js import fetch


async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    print(f'url={url}')
    print(f'method={method}')
    print(f'data=`{data}`')
    response = await fetch(url, method=method, body=data)
    text = await response.text()
    return text



import wwwpy.remote.component as wpc
import js

import logging

logger = logging.getLogger(__name__)


class Component1(wpc.Component, tag_name='component-1'):

    def init_component(self):
        # language=html
        self.element.innerHTML = """<span>component-1</span>"""
